-----------------------------------------------------------------------------------------------
--[[
    AlterFrame by tangjiasen

    It's like Skyrim with bars

    TODO:
    - Cast bars
    - Pet frames
    - Other assets

    If you are currently reading this, hello!
--]]
-----------------------------------------------------------------------------------------------
 
require "Window"
require "GameLib"
require "Apollo"
require "Unit"
require "AbilityBook"

local version = "0.8.3"

local change_log = string.format("AlterFrame v"..version.." Changes/fixes:\n%s%s%s%s",
                    "+ Target portrait assets", "\n+ Raid markers", 
                    "\n~ Fixed a minor lua bug related to logging into a character with no prior AlterFrame save data",
                    "\n~ Minor changes were made to the player frame; frame positions have been reset to apply these changes")
 
-----------------------------------------------------------------------------------------------
-- AlterFrame Module Definition
-----------------------------------------------------------------------------------------------
local AlterFrame = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local half_health_threshold     = .5
local critical_health_threshold = .3

local classes =
{
    [GameLib.CodeEnumClass.Warrior]         = {"IconSprites:Icon_Windows_UI_CRB_Warrior",       "Warrior"       },
    [GameLib.CodeEnumClass.Engineer]        = {"IconSprites:Icon_Windows_UI_CRB_Engineer",      "Engineer"      },
    [GameLib.CodeEnumClass.Esper]           = {"IconSprites:Icon_Windows_UI_CRB_Esper",         "Esper"         },
    [GameLib.CodeEnumClass.Medic]           = {"IconSprites:Icon_Windows_UI_CRB_Medic",         "Medic"         },
    [GameLib.CodeEnumClass.Stalker]         = {"IconSprites:Icon_Windows_UI_CRB_Stalker",       "Stalker"       },
    [GameLib.CodeEnumClass.Spellslinger]    = {"IconSprites:Icon_Windows_UI_CRB_Spellslinger",  "Spellslinger"  }
}

local rank_descriptions =
{
    [Unit.CodeEnumRank.Fodder]      =   {Apollo.GetString("TargetFrame_Fodder"),       Apollo.GetString("TargetFrame_VeryWeak"),      "spr_TargetFrame_ClassIcon_Fodder"    },
    [Unit.CodeEnumRank.Minion]      =   {Apollo.GetString("TargetFrame_Minion"),       Apollo.GetString("TargetFrame_Weak"),          "spr_TargetFrame_ClassIcon_Minion"    },
    [Unit.CodeEnumRank.Standard]    =   {Apollo.GetString("TargetFrame_Grunt"),        Apollo.GetString("TargetFrame_EasyAppend"),    "spr_TargetFrame_ClassIcon_Standard"  },
    [Unit.CodeEnumRank.Champion]    =   {Apollo.GetString("TargetFrame_Challenger"),   Apollo.GetString("TargetFrame_AlmostEqual"),   "spr_TargetFrame_ClassIcon_Champion"  },
    [Unit.CodeEnumRank.Superior]    =   {Apollo.GetString("TargetFrame_Superior"),     Apollo.GetString("TargetFrame_Strong"),        "spr_TargetFrame_ClassIcon_Superior"  },
    [Unit.CodeEnumRank.Elite]       =   {Apollo.GetString("TargetFrame_Prime"),        Apollo.GetString("TargetFrame_VeryStrong"),    "spr_TargetFrame_ClassIcon_Elite"     }
}

local difficulty_info =
{
    [-4] = {ApolloColor.new("ConTrivial"),     Apollo.GetString("TargetFrame_Trivial"),    Apollo.GetString("TargetFrame_NoXP"),               "ff7d7d7d"},
    [-3] = {ApolloColor.new("ConInferior"),    Apollo.GetString("TargetFrame_Inferior"),   Apollo.GetString("TargetFrame_VeryReducedXP"),      "ff01ff07"},
    [-2] = {ApolloColor.new("ConMinor"),       Apollo.GetString("TargetFrame_Minor"),      Apollo.GetString("TargetFrame_ReducedXP"),          "ff01fcff"},
    [-1] = {ApolloColor.new("ConEasy"),        Apollo.GetString("TargetFrame_Easy"),       Apollo.GetString("TargetFrame_SlightlyReducedXP"),  "ff597cff"},
     [0] = {ApolloColor.new("ConAverage"),     Apollo.GetString("TargetFrame_Average"),    Apollo.GetString("TargetFrame_StandardXP"),         "ffffffff"},
     [1] = {ApolloColor.new("ConModerate"),    Apollo.GetString("TargetFrame_Moderate"),   Apollo.GetString("TargetFrame_SlightlyMoreXP"),     "ffffff00"},
     [2] = {ApolloColor.new("ConTough"),       Apollo.GetString("TargetFrame_Tough"),      Apollo.GetString("TargetFrame_IncreasedXP"),        "ffff8000"},
     [3] = {ApolloColor.new("ConHard"),        Apollo.GetString("TargetFrame_Hard"),       Apollo.GetString("TargetFrame_HighlyIncreasedXP"),  "ffff0000"},
     [4] = {ApolloColor.new("ConImpossible"),  Apollo.GetString("TargetFrame_Impossible"), Apollo.GetString("TargetFrame_GreatlyIncreasedXP"), "ffff00ff"}
}

local raid_marker_icons =
{
    "Icon_Windows_UI_CRB_Marker_Bomb",
    "Icon_Windows_UI_CRB_Marker_Ghost",
    "Icon_Windows_UI_CRB_Marker_Mask",
    "Icon_Windows_UI_CRB_Marker_Octopus",
    "Icon_Windows_UI_CRB_Marker_Pig",
    "Icon_Windows_UI_CRB_Marker_Chicken",
    "Icon_Windows_UI_CRB_Marker_Toaster",
    "Icon_Windows_UI_CRB_Marker_UFO",
}

local disposition_colors =
{
    [Unit.CodeEnumDisposition.Neutral]  = {ApolloColor.new("DispositionNeutral"), "border_target_neutral"},
    [Unit.CodeEnumDisposition.Hostile]  = {ApolloColor.new("DispositionHostile"), "border_target_hostile"},
    [Unit.CodeEnumDisposition.Friendly] = {ApolloColor.new("DispositionFriendly"), "border_target_friendly"}
}

local segments = 
{
    [1]  = {"H1",  "S1",  "A1"},
    [2]  = {"H2",  "S2",  "A2"},
    [3]  = {"H3",  "S3",  "A3"},
    [4]  = {"H4",  "S4",  "A4"},
    [5]  = {"H5",  "S5",  "A5"},
    [6]  = {"H6",  "S6",  "A6"},
    [7]  = {"H7",  "S7",  "A7"},
    [8]  = {"H8",  "S8",  "A8"},
    [9]  = {"H9",  "S9",  "A9"},
    [10] = {"H10", "S10", "A10"},
    [11] = {"H11", "S11", "A11"},
    [12] = {"H12", "S12", "A12"},
    [13] = {"H13", "S13", "A13"},
    [14] = {"H14", "S14", "A14"},
    [15] = {"H15", "S15", "A15"},
    [16] = {"H16", "S16", "A16"},
    [17] = {"H17", "S17", "A17"},
    [18] = {"H18", "S18", "A18"},
    [19] = {"H19", "S19", "A19"}
}

local approved_unit_types = { "Player", "NonPlayer", "Pet", "Scanner", "Turret" }

local tooltip_body_color = "ffc0c0c0"
local tooltip_title_color = "ffdadada"

-----------------------------------------------------------------------------------------------
-- Variables
-----------------------------------------------------------------------------------------------
local flip_target_fill    = false
local show_percent        = false
local test_called         = false
local move_toggle         = false

local level_color         = ""
local focus_level_color   = ""
local rank_tooltip        = ""

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function AlterFrame:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self

    return o
end

function AlterFrame:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

function AlterFrame:OnSave(eLevel)
    local percent_enabled  = show_percent
    local target_fill_flip = flip_target_fill
    local ver              = version

    local player_left, player_top, player_right, player_bottom  = self.wndMain:GetAnchorOffsets()
    local target_left, target_top, target_right, target_bottom  = self.wndTarget:GetAnchorOffsets()
    local focus_left,  focus_top,  focus_right,  focus_bottom   = self.wndFocus:GetAnchorOffsets()

    save_data = { 
        version          = ver,
        show_percent     = percent_enabled,
        flip_target_fill = target_fill_flip,
        player_offsets   = {player_left, player_top, player_right, player_bottom},
        target_offsets   = {target_left, target_top, target_right, target_bottom},
        focus_offsets    = {focus_left,  focus_top,  focus_right,  focus_bottom}
    }

    if eLevel == GameLib.CodeEnumAddonSaveLevel.Character then
        return save_data
    elseif eLevel == GameLib.CodeEnumAddonSaveLevel.Account then
        self.account_data[self.strCharacterName] = save_data
        return self.account_data
    else
        return nil
    end
end

function AlterFrame:OnRestore(eLevel, tData)
    if eLevel == GameLib.CodeEnumAddonSaveLevel.Character then
        -- OnRestore is only used to load addon settings; these settings will be applied by OnGreeting
        self.saved_data = tData

    elseif eLevel == GameLib.CodeEnumAddonSaveLevel.Account then
        self.account_data = tData
    end
end

-----------------------------------------------------------------------------------------------
-- AlterFrame OnLoad
-----------------------------------------------------------------------------------------------
function AlterFrame:OnLoad()
    -- load our form files
	self.xmlDoc = XmlDoc.CreateFromFile("AlterFrame.xml")
    Apollo.LoadSprites("AlterFrameSprites.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)

    self.account_data = {}
end

-----------------------------------------------------------------------------------------------
-- AlterFrame OnDocLoaded
-----------------------------------------------------------------------------------------------
function AlterFrame:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "Player", "FixedHudStratum", self)
        self.wndTarget = Apollo.LoadForm(self.xmlDoc, "Target", "FixedHudStratum", self)
        self.wndFocus = Apollo.LoadForm(self.xmlDoc, "Focus", "FixedHudStratum", self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndMain:Show(true)

		self.xmlDoc = nil
		
        Apollo.RegisterEventHandler("VarChange_FrameCount", "OnFrame", self)
        Apollo.RegisterSlashCommand("focus", "OnFocusSlashCommand", self)
        Apollo.RegisterSlashCommand("alter", "OnInvoke", self)
        Apollo.RegisterTimerHandler("Greeting", "OnGreeting", self)

        Apollo.CreateTimer("Greeting", 1.5, false)
	end
end

-----------------------------------------------------------------------------------------------
-- AlterFrame Functions
-----------------------------------------------------------------------------------------------
function AlterFrame:OnFrame()
    self.unit_player = GameLib.GetPlayerUnit()

    if self.unit_player ~= nil then
        self.unit_target = self.unit_player:GetTarget()
        self.unit_focus  = self.unit_player:GetAlternateTarget()

        self:UpdatePlayerFrame()
        self:UpdateTargetFrame()
        self:UpdateFocusFrame()
    end
end

-----------------------------------------------------------------------------------------------
-- Update Functions
-----------------------------------------------------------------------------------------------
-- Main update function --
function AlterFrame:UpdateFrame(wndFrame, unit, is_player, is_target)
    local name              = unit:GetName()
    local class             = is_player and unit:GetClassId() or unit:GetRank()
    local level             = unit:GetLevel()
    local player_level      = self.unit_player:GetLevel()
    local disposition       = unit:GetDispositionTo(self.unit_player)
    local target            = unit:GetTarget()
    local in_combat         = unit:IsInCombat()

    local percent_div       = (1 / 19)
    local level_diff        = level ~= nil and level - player_level or -9001

    local max_hp            = unit:GetMaxHealth()
    local current_hp        = unit:GetHealth()
    local hp_percent        = current_hp / max_hp

    local cc_armor          = unit:GetInterruptArmorValue()
    local pvp_flagged       = GameLib.IsPvpFlagged()

    local max_shield        = unit:GetShieldCapacityMax()
    local current_shield    = unit:GetShieldCapacity()
    local shield_percent    = ((max_shield ~= 0 and current_shield <= max_shield) and current_shield / max_shield) or (max_shield ~= 0 and current_shield > max_shield and 1) or 0
    local shield_active = shield_percent > 0

    local max_absorb        = unit:GetAbsorptionMax()
    local current_absorb    = 0
    if max_absorb > 0 then current_absorb = unit:GetAbsorptionValue() end
    local absorb_percent    = max_absorb > 0 and current_absorb / max_hp or 0

    local tooltip           = string.format("HP: %s / %s (%d%s)%s%s", self:FormatNumber(current_hp), self:FormatNumber(max_hp), hp_percent * 100, "%", 
                    shield_percent > 0 and "\nShield: "..self:FormatNumber(current_shield).." / "..self:FormatNumber(max_shield).." ("..(shield_percent * 100).."%)" or "", 
                    current_absorb > 0 and "\nAbsorb: "..self:FormatNumber(current_absorb).." / "..self:FormatNumber(max_absorb).." ("..((current_absorb / max_absorb) * 100).."%)" or "")

    local sprite_to_use = (shield_active and hp_percent > 0) and "border_w_shield" or "border_wo_shield"

    local ranges            = {1, 19}
    local is_target_frame   = wndFrame == self.wndTarget
    local upper             = (is_target_frame and flip_target_fill) and ranges[1] or ranges[2]
    local lower             = (is_target_frame and flip_target_fill) and ranges[2] or ranges[1]
    local delta             = (is_target_frame and flip_target_fill) and -1 or 1

    --if wndFrame == self.wndMain then wndFrame:FindChild("SpecButton"):Show(not in_combat) end
    
    wndFrame:FindChild("Border"):SetSprite(sprite_to_use)
    wndFrame:FindChild("BG"):SetSprite((shield_percent > 0 and hp_percent > 0) and "bg_w_shield" or "bg_wo_shield")

    wndFrame:FindChild("HPLabel"):SetTextColor((current_absorb > 0 and "xkcdAmber") or (hp_percent >= 1 and "green") or (hp_percent <= half_health_threshold and hp_percent > critical_health_threshold and "yellow") or (hp_percent < critical_health_threshold and "red") or "white")
    wndFrame:FindChild("ShieldLabel"):SetTextColor((shield_percent >= 1 and "cyan") or (shield_percent <= half_health_threshold and shield_percent > critical_health_threshold and "yellow") or (shield_percent < critical_health_threshold and "red") or "white")

    wndFrame:FindChild("HPLabel"):SetText(not show_percent and self:FormatNumber(current_hp + current_absorb) or string.format("%d%s", hp_percent * 100, "%"))
    wndFrame:FindChild("ShieldLabel"):SetText((shield_percent > 0 and hp_percent > 0) and (not show_percent and self:FormatNumber(current_shield) or string.format("%d%s", shield_percent * 100, "%")) or "")

    -- Health, Shield, and Absorb --
    for idy = 1, 3 do
        for idx = lower, upper, delta do
            local idx_step = (is_target_frame and flip_target_fill) and (20 - idx) or idx

            local percent = (idy == 1 and hp_percent) or (idy == 2 and shield_percent) or absorb_percent

            local segment_wnd = wndFrame:FindChild(segments[idx][idy])

            if idy == 1 then
                if (percent >= (percent_div * idx_step) and (is_target_frame and flip_target_fill and idx < 19 or idx > 1)) or (percent > 0 and idx == ((is_target_frame and flip_target_fill) and 19 or 1)) then
                    if unit:IsInCCState(Unit.CodeEnumCCState.Vulnerability) and not is_player then
                        segment_wnd:SetSprite(idx % 2 == 1 and "down_vulnerable" or "up_vulnerable")
                    elseif hp_percent <= critical_health_threshold then
                        segment_wnd:SetSprite(idx % 2 == 1 and "down_critical" or "up_critical")
                    elseif hp_percent <= half_health_threshold and hp_percent > critical_health_threshold then
                        segment_wnd:SetSprite(idx % 2 == 1 and "down_half" or "up_half")
                    else
                        segment_wnd:SetSprite(idx % 2 == 1 and "down_health" or "up_health")
                    end
                    segment_wnd:Show(true)
                else
                    segment_wnd:SetSprite(idx % 2 == 1 and "down_white" or "up_white")
                    segment_wnd:Show(false)
                end
            else
                if (percent >= (percent_div * idx_step) and (is_target_frame and flip_target_fill and idx < 19 or idx > 1)) or (percent > 0 and idx == ((is_target_frame and flip_target_fill) and 19 or 1)) and hp_percent > 0 then
                    segment_wnd:SetSprite(idx % 2 == (idy == 2 and 0 or 1) and (idy == 2 and "down_shield" or "down_absorb") or (idy == 2 and "up_shield" or "up_absorb"))
                    segment_wnd:Show(true, idy == 3 and true or false)
                else
                    segment_wnd:SetSprite(idx % 2 == (idy == 2 and 0 or 1) and "down_white" or "up_white")
                    segment_wnd:Show(false)
                end
            end

        end
    end

    -- Border Dressing --
    local under_cc      = unit:IsInCCState(Unit.CodeEnumCCState.Stun) or unit:IsInCCState(Unit.CodeEnumCCState.Root) or unit:IsInCCState(Unit.CodeEnumCCState.Knockdown) or unit:IsInCCState(Unit.CodeEnumCCState.Vulnerability)
    local slowed_cc     = unit:IsInCCState(Unit.CodeEnumCCState.Snare) or unit:IsInCCState(Unit.CodeEnumCCState.Daze)

    wndFrame:FindChild("Critical"):SetSprite(sprite_to_use.."_critical")
    wndFrame:FindChild("Critical"):Show(hp_percent <= critical_health_threshold and hp_percent > 0)
    wndFrame:FindChild("Absorb"):SetSprite(sprite_to_use.."_absorb")
    wndFrame:FindChild("Absorb"):Show(current_absorb > 0, current_absorb > 0)
    wndFrame:FindChild("CC"):SetSprite(sprite_to_use.."_cc")
    wndFrame:FindChild("CC"):Show(under_cc, under_cc)
    wndFrame:FindChild("Slowed"):Show(slowed_cc, slowed_cc)

    wndFrame:FindChild("Border"):SetTooltip(tooltip)

    -- Interrupt Armor --
    wndFrame:FindChild("InterruptArmor"):Show(cc_armor > 0, cc_armor > 0)
    wndFrame:FindChild("CCArmor"):SetText(cc_armor)

    -- PvP Flag --
    wndFrame:FindChild("PvPFlag"):Show(pvp_flagged)

    -- Buff/Harm Bars --
    wndFrame:FindChild("BeneBuffBar"):SetUnit(unit)
    wndFrame:FindChild("HarmBuffBar"):SetUnit(unit)

    -- Level, Name, Portrait and Class --
    wndFrame:FindChild("Level"):SetText(level ~= nil and level or "--")

    if is_target then
        if level_diff >= 4 then
            level_color = difficulty_info[4][1]
            self.idx = 4
        elseif level_diff <= -4 then
            level_color = difficulty_info[-4][1]
            self.idx = -4
        else
            level_color = difficulty_info[level_diff][1]
            self.idx = level_diff
        end

        -- Target and ToT Portrait --
        wndFrame:FindChild("TargetPortrait"):SetCostume(unit)
        wndFrame:FindChild("TargetPortrait"):SetTooltip(name)
        wndFrame:FindChild("TargetDisposition"):SetSprite(disposition_colors[disposition][2])

        wndFrame:FindChild("Level"):SetTextColor(level_color)
        wndFrame:FindChild("Name"):SetText(name)
        wndFrame:FindChild("Name"):SetTextColor(disposition_colors[disposition][1])

        if target ~= nil and target:GetMaxHealth() ~= nil then
            wndFrame:FindChild("ToTPortrait"):SetCostume(target)
            wndFrame:FindChild("ToTPortrait"):SetTooltip(target:GetName())
            wndFrame:FindChild("TotDisposition"):SetSprite(disposition_colors[target:GetDispositionTo(self.unit_player)][2])
            wndFrame:FindChild("ToTP"):Show(true)
        else
            wndFrame:FindChild("ToTP"):Show(false)
        end

        -- XP reward tooltip --
        local xp_reward     = String_GetWeaselString(Apollo.GetString("TargetFrame_TargetXPReward"), difficulty_info[self.idx][3])
        local level_tooltip = self:HelperBuildTooltip(xp_reward, difficulty_info[self.idx][2], difficulty_info[self.idx][4])

        wndFrame:FindChild("Level"):SetTooltip(level_tooltip)
    end

    -- Class/Mob Rank Icon --
    local class_wnd = wndFrame:FindChild("Class")
    if is_player then
        class_wnd:SetSprite(classes[class][1])
        wndFrame:FindChild("ClassBG"):SetSprite(classes[class][1])
        class_wnd:SetTooltip(classes[class][2])
    else
        if rank_descriptions[class] ~= nil then
            local rank_desc = String_GetWeaselString(Apollo.GetString("TargetFrame_CreatureRank"), rank_descriptions[class][2])
            rank_tooltip    = self:HelperBuildTooltip(rank_desc, rank_descriptions[class][1])
        end

        class_wnd:SetSprite(rank_descriptions[class][3])
        wndFrame:FindChild("ClassBG"):SetSprite(rank_descriptions[class][3])
        class_wnd:SetTooltip(rank_tooltip)
    end

    -- Raid Marker --
    local marker_id = unit and unit:GetTargetMarker() or 0
    local raid_marker_wnd   = wndFrame:FindChild("RaidMarker")
    --local raid_marker_wndbg = wndFrame:FindChild("RaidMarkerBG")

    if unit and marker_id ~= 0 then
        raid_marker_wnd:SetSprite(raid_marker_icons[marker_id])
        raid_marker_wnd:Show(true)
    else
        raid_marker_wnd:SetSprite("")
        raid_marker_wnd:Show(false)
    end
end

-- Player frame update helper --
function AlterFrame:UpdatePlayerFrame()
    self:UpdateFrame(self.wndMain, self.unit_player, true, false)
end

-- Target frame update helper --
function AlterFrame:UpdateTargetFrame(unit)
    local valid_target = self:ValidateTarget()

    if valid_target then
        self.wndTarget:Show(true)
    else
        self.wndTarget:Show(false)
        return
    end

    self:UpdateFrame(self.wndTarget, self.unit_target, self.unit_target:IsACharacter(), true)
end

-- Focus frame update helper --
function AlterFrame:UpdateFocusFrame()
    if self.unit_focus ~= nil and self.unit_focus:GetMaxHealth() ~= nil then
        self.wndFocus:Show(true)
    else
        self.wndFocus:Show(false)
        return
    end

    self:UpdateFrame(self.wndFocus, self.unit_focus, self.unit_focus:IsACharacter(), true)
end

-----------------------------------------------------------------------------------------------
-- Misc Helper Functions
-----------------------------------------------------------------------------------------------
function AlterFrame:FormatNumber(number)
    function round(num, idp)
        local mult = 10^(idp or 0)
        return math.floor(num * mult + 0.5) / mult
    end

    local millions        = number / 1000000
    local thousands       = math.floor(number / 1000)
    local hundreds        = number % 1000

    local formated_number = ""

    if millions >= 1 then
        formated_number = round(millions, 1).."m"
    elseif number / 100000 >= 1 then
        formated_number = round(number / 1000, 1).."k"
    else
        if number > 1000 and hundreds < 10 then
            formated_number = thousands..",00"..hundreds
        elseif number > 1000 and hundreds < 100 then
            formated_number = thousands..",0"..hundreds
        elseif number > 1000 then
            formated_number = thousands..","..hundreds
        else
            formated_number = number
        end
    end

    return formated_number
end

function AlterFrame:ValidateTarget()
    local valid_target = false
    if self.unit_target ~= nil and self.unit_target:GetMaxHealth() ~= nil then
        for k, v in pairs(approved_unit_types) do
            if self.unit_target:GetType() == v then
                valid_target = true
                break
            else
                valid_target = false
            end
        end
    end

    return valid_target
end

function AlterFrame:ChangeSpec()
    AbilityBook.NextSpec()
end

function AlterFrame:OnGenerateTooltip(wndHandler, wndControl, tType, splBuff)
    if wndHandler == wndControl then
        return
    end
    Tooltip.GetBuffTooltipForm(self, wndControl, splBuff)
end

function AlterFrame:HelperBuildTooltip(body, title, title_color)
    if body == nil then return end
    local tooltip = string.format("<T Font=\"CRB_InterfaceMedium\" TextColor=\"%s\">%s</T>", tooltip_body_color, body)
    if title ~= nil then -- if a title has been passed, add it (optional)
        tooltip = string.format("<P>%s</P>", tooltip)
        local title = string.format("<P Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">%s</P>", title_color or tooltip_title_color, title)
        tooltip = title..tooltip
    end
    return tooltip
end

function AlterFrame:OnMouseButtonDown(wndHandler, wndControl, eMouseButton, x, y)
    if eMouseButton == GameLib.CodeEnumInputMouse.Left then
        if wndHandler == self.wndMain then
            GameLib.SetTargetUnit(self.unit_player)
        elseif wndHandler == self.wndFocus then
            GameLib.SetTargetUnit(self.unit_focus)
        end
    end

    if eMouseButton == GameLib.CodeEnumInputMouse.Right and self:ValidateTarget() and wndHandler == self.wndTarget then
        Event_FireGenericEvent("GenericEvent_NewContextMenuPlayerDetailed", nil, self.unit_target:GetName(), self.unit_target)
        return true
    elseif eMouseButton == GameLib.CodeEnumInputMouse.Right and wndHandler == self.wndMain and (wndControl == self.wndMain:FindChild("BG") or wndControl == self.wndMain:FindChild("Border")) then
        Event_FireGenericEvent("GenericEvent_NewContextMenuPlayerDetailed", nil, self.unit_player:GetName(), self.unit_player)
        return true
    elseif eMouseButton == GameLib.CodeEnumInputMouse.Right and wndHandler == self.wndFocus then
        Event_FireGenericEvent("GenericEvent_NewContextMenuPlayerDetailed", nil, self.unit_focus:GetName(), self.unit_focus)
        return true
    end

    return false
end

-----------------------------------------------------------------------------------------------
-- Slash command functions
-----------------------------------------------------------------------------------------------
function AlterFrame:TogglePercent()
    show_percent = not show_percent
    Print("AlterFrame: percent "..(show_percent and "ON" or "OFF"))
end

function AlterFrame:ToggleFlip()
    flip_target_fill = not flip_target_fill
    Print("AlterFrame: target fill flipped "..(flip_target_fill and "ON" or "OFF"))
end

function AlterFrame:ToggleMove()
    move_toggle = not move_toggle
    self.wndMain:SetStyle("Moveable", move_toggle)
    self.wndTarget:SetStyle("Moveable", move_toggle)
    self.wndFocus:SetStyle("Moveable", move_toggle)

    Print("AlterFrame: move toggle "..(move_toggle and "ON" or "OFF"))
end

function AlterFrame:ShowSettings()
    local settings = {
        percent = show_percent,
        flip = flip_target_fill,
        move = move_toggle
    }

    Print("AlterFrame Settings")
    Print("---------------------")
    for k,v in pairs(settings) do
        Print(k.. " : "..(v and "ON" or "OFF"))
    end
end

function AlterFrame:ResetPosition()
    self.wndMain:SetAnchorOffsets(-154, 20, 154, 120)
    self.wndTarget:SetAnchorOffsets(-233, 215, 147, 335)
    self.wndFocus:SetAnchorOffsets(-147, 215, 233, 335)

    Print("AlterFrame: frame positions reset.")
end

function AlterFrame:OnInvoke(cmd, args)
    local valid_commands = { "percent", "flip", "settings", "move", "reset", "" }
    local setting_descriptions = {
        percent  = "Toggles percent view.",
        flip     = "Flips fill direction. (target frame only)",
        move     = "Unlocks frames to move.",
        settings = "Shows AlterFrame's current settings.",
        reset    = "Resets the position of all frames."
    }

    --Print(cmd.." | "..args)
    for k, v in pairs(valid_commands) do
        if string.lower(args) == v then
            break
        elseif k == #valid_commands then
            Print("AlterFrame: \'"..args.."\' is not a valid command.")
            return
        end
    end

    if string.lower(args) == "percent" then
        self:TogglePercent()
    elseif string.lower(args) == "flip" then
        self:ToggleFlip()
    elseif string.lower(args) == "settings" then
        self:ShowSettings()
    elseif string.lower(args) == "move" then
        self:ToggleMove()
    elseif string.lower(args) == "reset" then
        self:ResetPosition()
    else
        Print("AlterFrame v"..(version).." Commands")
        Print("Invoke a command with \"/alter <command>\"")
        Print("-----------------------------")
        for k, v in pairs(setting_descriptions) do
            Print(k.." : "..v)
        end
    end
end

function AlterFrame:OnFocusSlashCommand()
    self.unit_player:SetAlternateTarget(self.unit_target)
end

-----------------------------------------------------------------------------------------------
-- Timer functions
-----------------------------------------------------------------------------------------------
function AlterFrame:OnGreeting()
    if self.saved_data ~= nil then
        -- Reload frame positions
        if self.saved_data.version == version then
            if self.saved_data.player_offsets   ~= nil then self.wndMain:SetAnchorOffsets(unpack(self.saved_data.player_offsets))   end
            if self.saved_data.target_offsets   ~= nil then self.wndTarget:SetAnchorOffsets(unpack(self.saved_data.target_offsets)) end
            if self.saved_data.focus_offsets    ~= nil then self.wndFocus:SetAnchorOffsets(unpack(self.saved_data.focus_offsets))   end
        else
            Print(change_log)
        end

        -- Reload other settings
        if self.saved_data.show_percent     ~= nil then show_percent     = self.saved_data.show_percent     end
        if self.saved_data.flip_target_fill ~= nil then flip_target_fill = self.saved_data.flip_target_fill end
    end

    Print("AlterFrame v"..(version).." ready")
    Print("Type /alter for commands.")
    Apollo.StopTimer("Greeting")
end

-----------------------------------------------------------------------------------------------
-- AlterFrame Instance
-----------------------------------------------------------------------------------------------
local AlterFrameInst = AlterFrame:new()
AlterFrameInst:Init()
