-----------------------------------------------------------------------------------------------
-- Client Lua Script for AlterSprint
-- Copyright (c) tangjiasen. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "GameLib"
require "Apollo"
 
-----------------------------------------------------------------------------------------------
-- AlterSprint Module Definition
-----------------------------------------------------------------------------------------------
local AlterSprint = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local segments = 
{
    [1] = { "D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9" },
    [2] = { "D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8" }
}
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function AlterSprint:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self
    
    self.just_filled = false

    return o
end

function AlterSprint:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- AlterSprint OnLoad
-----------------------------------------------------------------------------------------------
function AlterSprint:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("AlterSprint.xml")
    Apollo.LoadSprites("AlterSprintSprites.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- AlterSprint OnDocLoaded
-----------------------------------------------------------------------------------------------
function AlterSprint:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "SprintForm", "FixedHudStratumLow", self)
		
	    self.wndMain:Show(false, true)

		self.xmlDoc = nil

        Apollo.RegisterTimerHandler("SprintMeterGracePeriod",   "OnSprintMeterGracePeriod", self)
        Apollo.RegisterEventHandler("VarChange_FrameCount",     "OnFrame", self)

	end
end

-----------------------------------------------------------------------------------------------
-- AlterSprint Functions
-----------------------------------------------------------------------------------------------
function AlterSprint:OnFrame()
    local unit_player = GameLib.GetPlayerUnit()
    if not unit_player then return end

    local window_visible        = self.wndMain:IsVisible()

    local current_endurance     = unit_player:GetResource(0)
    local max_endurance         = unit_player:GetMaxResource(0)
    local endurance_percent     = current_endurance / max_endurance
    local endurance_div         = (1 / 9)

    local current_dash          = unit_player:GetResource(7)
    local max_dash              = unit_player:GetMaxResource(7)
    local dash_percent          = current_dash / max_dash
    local dash_div              = (1 / 8)

    local at_max = (current_endurance == max_endurance and current_dash == max_dash) or unit_player:IsDead()

    local bars = { "Sprint", "Dash" }

    for idy = 1, 2 do
        if idy == 1 then
            for idx = 1, 9 do
                local sprite_to_use = (endurance_percent > .5 and "_blue") or ((endurance_percent <= .5 and endurance_percent > .25) and "_yellow") or "_red"
                if endurance_percent >= (endurance_div * idx) then
                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):SetSprite(idx % 2 == 1 and "up"..sprite_to_use or "down"..sprite_to_use)
                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):Show(true)
                else
                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):SetSprite(idx % 2 == 1 and "up_white" or "down_white")
                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):Show(false)
                end
            end
        else
            for idx = 1, 8 do
                if dash_percent >= (dash_div * idx) then
                    local dash1_sprite = dash_percent >= .5 and "_green" or "_red"
                    local dash2_sprite = dash_percent >= 1 and "_green" or "_red"

                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):SetSprite(idx % 2 == 1 and (idx <= 4 and "up"..dash1_sprite or "down"..dash2_sprite) or (idx <= 4 and "down"..dash1_sprite or "up"..dash2_sprite))
                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):Show(true)
                else
                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):SetSprite(idx % 2 == 1 and (idx <= 4 and "up_white" or "down_white") or (idx <= 4 and "down_white" or "up_white"))
                    self.wndMain:FindChild("SprintForm:"..bars[idy]..":"..segments[idy][idx]):Show(false)
                end
            end
        end
    end

    if window_visible and at_max and not self.just_filled then
        self.just_filled = true
        Apollo.StopTimer("SprintMeterGracePeriod")
        Apollo.CreateTimer("SprintMeterGracePeriod", .5, false)
    end

    if not at_max then
        self.just_filled = false
        Apollo.StopTimer("SprintMeterGracePeriod")
    end

    self.wndMain:Show(not at_max or self.just_filled)
end


function AlterSprint:OnSprintMeterGracePeriod()
    Apollo.StopTimer("SprintMeterGracePeriod")
    self.just_filled = false
    if self.wndMain and self.wndMain:IsValid() then
        self.wndMain:Show(false)
    end
end

-----------------------------------------------------------------------------------------------
-- AlterSprint Instance
-----------------------------------------------------------------------------------------------
local AlterSprintInst = AlterSprint:new()
AlterSprintInst:Init()
